#include <assert.h>
#include <math.h>

#include "paths/paths.h"
#include "paths/output_paths.h"
#include "actions/cats_actions.h"
#include "misc/cats_random.h"
#include "data/cats_grid.h"
#include "threading/threading-helpers.h"
#include "modules/module_header.h"

#include "butterflies_populations.h"
#include "butterflies_main.h"
#include "butterflies_inline.h"
#include "inline_population.h"
#include "inline.h"
#include "populations/population.h"


cats_dt_rates bf_egg_to_adult_survival_rate(cats_dt_rates reproduction_rate, cats_dt_rates eggs, cats_dt_rates female_fraction)
{
        return (reproduction_rate / female_fraction) / eggs;
}

void bf_cell_maturation(struct cats_grid *grid, struct cats_thread_info *ts, cats_dt_coord row, cats_dt_coord col,
                        bool check_exclusion)
{

        const struct cats_configuration *conf = ts->conf;
        assert (conf->grid_count == 1);

        if (check_exclusion) {
                if (bf_cell_excluded_by_generation(grid, row, col)) return;
        }



        const int module_id = CATS_MODULE_ID;
        struct grid_data_butterflies *data = grid->grid_modules[module_id].module_data;

        if (data->eggs[row][col] == 0) return;
        if (data->eggs[row][col] < 0) {
                log_message(LOG_ERROR, "Number of eggs < 0: row %d col %d: %f", row, col, data->eggs[row][col]);
                exit(EXIT_FAILURE);
        }

        //int orig_eggs = data->eggs[row][col];

        // the number of generations per cell is usually not an integer value
        // the non-integer part of the number of generations is used in the generation which is
        // one greater than the integer part of the number of generations
        // i.e. if the number of generations is 2.4, 40% of all eggs in the cell
        // will be considered as if they belonged to generation 3
        // the minimum number of generations is 1 if the cell is climatically viable (and the host species is present)
        // if the number of generations is 0, the cell will be ignored
        //
        // note that the generations are processed backwards
        // i.e. if the maximum number of generations is 5, individual cells will have 1 ... 5 generations
        // first the all cells which have 5 (> 4) generations are modelled
        // then all the cells with have at least 4 (> 3) generations are modelled
        // ...
        // at last all cells with have at least 1 generation are modelled

        float this_generation_fraction = 1.0f;
        int32_t local_max_generation = (int32_t) ceilf(data->generations[row][col]);

        if (data->generation_current == local_max_generation) {
                this_generation_fraction = (float) local_max_generation - data->generations[row][col];
                assert(this_generation_fraction >= 0);
                assert(this_generation_fraction <= 1.0);
        }

        //cats_dt_population max_cc = (cats_dt_population) grid->param.carrying_capacity.max_rate;
        float eggs = this_generation_fraction * data->eggs[row][col];

        if (eggs > data->eggs[row][col]) {
                log_message(LOG_ERROR, "Removing more eggs than present: %d %d: %f/%f", row, col, data->eggs[row][col],
                            eggs);
                exit_cats(EXIT_FAILURE);
        }

        data->eggs[row][col] -= eggs;


        assert(data->eggs[row][col] >= 0);
        if (eggs == 0) return;

        // not capped, we can have more adults than CC
        struct conf_data_butterflies *module_conf = CATS_MODULE_DATA;
        cats_dt_environment suit = get_suitability(grid, row, col);
        cats_dt_rates reproduction_rate = calculate_rate(&module_conf->reproduction_rate, NAN, &grid->param,
                                                         grid, row, col, NULL);
        if (suit < grid->param.OT && reproduction_rate > 0) {
                log_message(LOG_ERROR, "Suitability %f under OT %Lf, but adults per female = %Lf", suit, grid->param.OT, reproduction_rate);
                exit_cats(EXIT_FAILURE);
        }

        printf("row %d col %d: suitability: %f, eggs: %f, reproduction rate %Lf\n", row, col, suit, eggs, reproduction_rate);

        if (reproduction_rate == 0) {
                data->eggs[row][col] = 0;
                set_population_ignore_cc(grid, row, col, 0);
                return;
        }




        cats_dt_rates survival = bf_egg_to_adult_survival_rate(reproduction_rate, eggs, module_conf->female_fraction);
        cats_dt_population adults = poisson(ts->rng, eggs * survival);
        assert(adults >= 0);


        set_population(grid, row, col, adults);

}


void bf_area_kill_adults(struct cats_grid *grid, struct cats_thread_info *ts)
{
        const cats_dt_coord start_row = ts->area.start_row;
        const cats_dt_coord end_row = ts->area.end_row;
        const cats_dt_coord start_col = ts->area.start_col;
        const cats_dt_coord end_col = ts->area.end_col;

        for (cats_dt_coord row = start_row; row < end_row; row++) {
                for (cats_dt_coord col = start_col; col < end_col; col++) {
                        set_population_ignore_cc(grid, row, col, 0);
                }
        }
}